@@ -14,6 +14,8 @@
 import mock
 
 from keystoneauth1 import exceptions as ks_exc
+from keystoneauth1 import loading as keystone
+from oslo_serialization import jsonutils
 from oslo_utils import uuidutils
 
 from neutron_lib._i18n import _
@@ -39,6 +41,65 @@ INVENTORY = {
 }
 
 
+class TestNoAuthClient(base.BaseTestCase):
+
+    def setUp(self):
+        super(TestNoAuthClient, self).setUp()
+        self.noauth_client = place_client.NoAuthClient('placement/')
+        self.body_json = jsonutils.dumps({'name': 'foo'})
+        self.uuid = '42'
+
+    @mock.patch.object(place_client.NoAuthClient, 'request')
+    def test_get(self, mock_request):
+        self.noauth_client.get('resource_providers', '')
+        mock_request.assert_called_with('placement/resource_providers', 'GET')
+
+    @mock.patch.object(place_client.NoAuthClient, 'request')
+    def test_post(self, mock_request):
+        self.noauth_client.post('resource_providers', self.body_json, '')
+        mock_request.assert_called_with('placement/resource_providers', 'POST',
+                                        body=self.body_json)
+
+    @mock.patch.object(place_client.NoAuthClient, 'request')
+    def test_put(self, mock_request):
+        self.noauth_client.put('resource_providers/%s' % self.uuid,
+                               self.body_json, '')
+        mock_request.assert_called_with(
+            'placement/resource_providers/%s' % self.uuid, 'PUT',
+            body=self.body_json)
+
+    @mock.patch.object(place_client.NoAuthClient, 'request')
+    def test_delete(self, mock_request):
+        self.noauth_client.delete('resource_providers/%s' % self.uuid, '')
+        mock_request.assert_called_with(
+            'placement/resource_providers/%s' % self.uuid, 'DELETE')
+
+
+class TestPlacementAPIClientNoAuth(base.BaseTestCase):
+    def setUp(self):
+        super(TestPlacementAPIClientNoAuth, self).setUp()
+        self.config = mock.Mock()
+
+    @mock.patch('neutron_lib.placement.client.NoAuthClient', autospec=True)
+    def test__create_client_noauth(self, mock_auth_client):
+        self.config.placement.auth_type = 'noauth'
+        self.config.placement.auth_section = 'placement/'
+        self.placement_api_client = place_client.PlacementAPIClient(
+            self.config)
+        self.placement_api_client._create_client()
+        mock_auth_client.assert_called_once_with('placement/')
+
+    @mock.patch.object(keystone, 'load_auth_from_conf_options')
+    @mock.patch.object(keystone, 'load_session_from_conf_options')
+    def test__create_client(self, mock_session_from_conf, mock_auth_from_conf):
+        self.config.placement.auth_type = 'password'
+        self.placement_api_client = place_client.PlacementAPIClient(
+            self.config)
+        self.placement_api_client._create_client()
+        mock_auth_from_conf.assert_called_once_with(self.config, 'placement')
+        mock_session_from_conf.assert_called_once()
+
+
 class TestPlacementAPIClient(base.BaseTestCase):
 
     def setUp(self):
@@ -259,14 +320,6 @@ class TestPlacementAPIClient(base.BaseTestCase):
              'rc_name': RESOURCE_CLASS_NAME},
             expected_body)
 
-    def test_update_resource_inventory_inventory_conflict_exception(self):
-        self.placement_fixture.mock_put.side_effect = ks_exc.Conflict()
-        self.assertRaises(
-            n_exc.PlacementInventoryUpdateConflict,
-            self.placement_api_client.update_resource_provider_inventory,
-            RESOURCE_PROVIDER_UUID, INVENTORY,
-            RESOURCE_CLASS_NAME, resource_provider_generation=1)
-
     def test_update_resource_provider_inventory_not_found(self):
         # Test the resource provider not found case
         self.placement_fixture.mock_put.side_effect = ks_exc.NotFound(
@@ -440,3 +493,208 @@ class TestPlacementAPIClient(base.BaseTestCase):
             self.placement_api_client.delete_resource_class,
             RESOURCE_CLASS_NAME
         )
+
+    def test_update_rp_traits_caller_handles_generation_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = ks_exc.Conflict(
+            response=mock_resp)
+        self.assertRaises(
+            n_exc.PlacementResourceProviderGenerationConflict,
+            self.placement_api_client.update_resource_provider_traits,
+            resource_provider_uuid='resource provider uuid',
+            traits=['trait a', 'trait b'],
+            resource_provider_generation=3,
+        )
+        self.placement_fixture.mock_put.assert_called_once()
+
+    def test_update_rp_traits_callee_handles_generation_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = [
+            ks_exc.Conflict(response=mock_resp),
+            mock.Mock(),
+        ]
+        self.placement_api_client.update_resource_provider_traits(
+            resource_provider_uuid='resource provider uuid',
+            traits=['trait a', 'trait b'],
+            resource_provider_generation=None,
+        )
+        self.assertEqual(2, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_traits_reached_max_tries(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = 10 * [
+            ks_exc.Conflict(response=mock_resp),
+        ]
+        self.assertRaises(
+            n_exc.PlacementResourceProviderGenerationConflict,
+            self.placement_api_client.update_resource_provider_traits,
+            resource_provider_uuid='resource provider uuid',
+            traits=['trait a', 'trait b'],
+            resource_provider_generation=None,
+        )
+        self.assertEqual(10, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_traits_raise_other_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'some_other_code'}]}
+        self.placement_fixture.mock_put.side_effect = [
+            ks_exc.Conflict(response=mock_resp),
+            mock.Mock(),
+        ]
+        self.assertRaises(
+            n_exc.PlacementClientError,
+            self.placement_api_client.update_resource_provider_traits,
+            resource_provider_uuid='resource provider uuid',
+            traits=[],
+            resource_provider_generation=None,
+        )
+        self.assertEqual(1, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_inventory_caller_handles_generation_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = ks_exc.Conflict(
+            response=mock_resp)
+        self.assertRaises(
+            n_exc.PlacementResourceProviderGenerationConflict,
+            self.placement_api_client.update_resource_provider_inventory,
+            resource_provider_uuid='resource provider uuid',
+            inventory={},
+            resource_class='a resource class',
+            resource_provider_generation=3,
+        )
+        self.placement_fixture.mock_put.assert_called_once()
+
+    def test_update_rp_inventory_callee_handles_generation_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = [
+            ks_exc.Conflict(response=mock_resp),
+            mock.Mock(),
+        ]
+        self.placement_api_client.update_resource_provider_inventory(
+            resource_provider_uuid='resource provider uuid',
+            inventory={},
+            resource_class='a resource class',
+            resource_provider_generation=None,
+        )
+        self.assertEqual(2, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_inventory_reached_max_tries(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = 10 * [
+            ks_exc.Conflict(response=mock_resp),
+        ]
+        self.assertRaises(
+            n_exc.PlacementResourceProviderGenerationConflict,
+            self.placement_api_client.update_resource_provider_inventory,
+            resource_provider_uuid='resource provider uuid',
+            inventory={},
+            resource_class='a resource class',
+            resource_provider_generation=None,
+        )
+        self.assertEqual(10, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_inventory_raise_other_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'some_other_code'}]}
+        self.placement_fixture.mock_put.side_effect = [
+            ks_exc.Conflict(response=mock_resp),
+            mock.Mock(),
+        ]
+        self.assertRaises(
+            n_exc.PlacementClientError,
+            self.placement_api_client.update_resource_provider_inventory,
+            resource_provider_uuid='resource provider uuid',
+            inventory={},
+            resource_class='a resource class',
+            resource_provider_generation=None,
+        )
+        self.assertEqual(1, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_inventories_caller_handles_generation_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = ks_exc.Conflict(
+            response=mock_resp)
+        self.assertRaises(
+            n_exc.PlacementResourceProviderGenerationConflict,
+            self.placement_api_client.update_resource_provider_inventories,
+            resource_provider_uuid='resource provider uuid',
+            inventories={},
+            resource_provider_generation=3,
+        )
+        self.placement_fixture.mock_put.assert_called_once()
+
+    def test_update_rp_inventories_callee_handles_generation_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = [
+            ks_exc.Conflict(response=mock_resp),
+            mock.Mock(),
+        ]
+        self.placement_api_client.update_resource_provider_inventories(
+            resource_provider_uuid='resource provider uuid',
+            inventories={},
+            resource_provider_generation=None,
+        )
+        self.assertEqual(2, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_inventories_reached_max_tries(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'placement.concurrent_update'}]}
+        self.placement_fixture.mock_put.side_effect = 10 * [
+            ks_exc.Conflict(response=mock_resp),
+        ]
+        self.assertRaises(
+            n_exc.PlacementResourceProviderGenerationConflict,
+            self.placement_api_client.update_resource_provider_inventories,
+            resource_provider_uuid='resource provider uuid',
+            inventories={},
+            resource_provider_generation=None,
+        )
+        self.assertEqual(10, self.placement_fixture.mock_put.call_count)
+
+    def test_update_rp_inventories_raise_other_conflict(self):
+        mock_resp = mock.Mock()
+        mock_resp.text = ''
+        mock_resp.json = lambda: {
+            'errors': [{'code': 'some_other_code'}]}
+        self.placement_fixture.mock_put.side_effect = [
+            ks_exc.Conflict(response=mock_resp),
+            mock.Mock(),
+        ]
+        self.assertRaises(
+            n_exc.PlacementClientError,
+            self.placement_api_client.update_resource_provider_inventories,
+            resource_provider_uuid='resource provider uuid',
+            inventories={},
+            resource_provider_generation=None,
+        )
+        self.assertEqual(1, self.placement_fixture.mock_put.call_count)
